(*
  New Edit Window for Embedded Designer plugin

  Copyright (c) 2015 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit UNewedwin;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Rtti, System.RegularExpressions,
  Vcl.Controls, Vcl.Forms, Vcl.ActnList, Vcl.Tabs, ToolsAPI, Events;

procedure Register;

implementation

const
  sEditorViewsChangedEvent = '@Editorform@evEditorViewsChangedEvent';

var
  Org_ecNewEditWindowExecute: TNotifyEvent;
  ecNewEditWindow: TAction;
  evEditorViewsChangedEvent: ^TEvent;

function OpenNewEditWindow(AEditWindow: TCustomForm): Boolean;
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
  meth: TRttiMethod;
  fld: TRttiField;
  v: TValue;

  editView: TObject;
  mon: TMonitor;
  newEditWindow: TObject;
  r: TRect;
  viewBar: TTabSet;
begin
  Result := False;

  typ := ctx.GetType(AEditWindow.ClassType);
  if typ = nil then Exit;

  prop := typ.GetProperty('CurrentEditView');
  if prop = nil then Exit;
  editView := prop.GetValue(AEditWIndow).AsObject;
  if editView = nil then Exit;

  r := TControl(AEditWindow).BoundsRect;
  OffsetRect(r, GetSystemMetrics(SM_CXVSCROLL), GetSystemMetrics(SM_CYCAPTION));
  mon := TCustomForm(AEditWindow).Monitor;
  if mon <> nil then
  begin
    if r.Bottom > mon.Top + mon.Height then
      r.Bottom := mon.Top + mon.Height;

    if r.Right > mon.Left + mon.Width then
      r.Right := mon.Left + mon.Width;
  end;

  meth := typ.GetMethod('Create');
  if meth = nil then Exit;
  v := meth.Invoke(AEditWindow.ClassType, [Application]);
  if v.IsEmpty then Exit;
  newEditWindow := v.AsObject;
  if newEditWindow = nil then Exit;
  try
    meth := ctx.GetType(editView.ClassType).GetMethod('CloneEditView');
    if meth = nil then Exit;
    v := meth.Invoke(editView, [newEditWindow]);
    if v.IsEmpty then Exit;
    editView := v.AsObject;
    if editView = nil then Exit;

    meth := ctx.GetType(newEditWindow.ClassType).GetMethod('InsertView');
    if meth = nil then Exit;
    meth.Invoke(newEditWindow, [editView, '']);
    TControl(newEditWindow).BoundsRect := r;
    TControl(newEditWindow).Show;

    fld := typ.GetField('ViewBar');
    if fld = nil then Exit;
    viewBar := TTabSet(fld.GetValue(newEditWindow).AsObject);
    if viewBar = nil then Exit;
    viewBar.TabIndex := 0;

    Result := True;
  except
    newEditWindow.Free;
  end;
end;

procedure AltNewEditWindowExecute(Self, Sender: TObject);
var
  editorService: IOTAEditorServices;
  editView: IOTAEditView;
  editWindow: INTAEditWindow;
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
  v: TValue;
begin
  try
    if not Supports(BorlandIDEServices, IOTAEditorServices, editorService) then Abort;
    editView := editorService.TopView;
    if editView = nil then Abort;
    editWindow := editView.GetEditWindow;
    if editWindow = nil then Abort;
    if editWindow.Form = nil then Abort;

    typ := ctx.GetType(editWindow.Form.ClassType);
    if typ = nil then Abort;
    prop := typ.GetProperty('CanCloneCurEditView');
    if prop = nil then Abort;
    v := prop.GetValue(editWindow.Form);
    if v.IsEmpty then Abort;

    if v.AsBoolean then Abort;
    if not OpenNewEditWindow(editWindow.Form) then Abort;
  except
    if Assigned(Org_ecNewEditWindowExecute) then
      Org_ecNewEditWindowExecute(Sender);
  end;
end;

procedure MyEditorViewsChanged(Self, Sender: TObject; NewTabIndex: Integer; NewViewIndex: Integer);
var
  i: Integer;
  viewBar: TTabSet;
  ctx: TRttiContext;
  typ: TRttiType;
  fldViewBar: TRttiField;
begin
  if Sender = nil then Exit;
  if not SameText(Sender.ClassName, 'TEditWindow') then Exit;
  if string(TComponent(Sender).Name).EndsWith('_0') then Exit;

  typ := ctx.GetType(Sender.ClassType);
  if typ = nil then Exit;
  fldViewBar := typ.GetField('ViewBar');
  if fldViewBar = nil then Exit;
  viewBar := TTabSet(fldViewBar.GetValue(Sender).AsObject);
  for i := viewBar.Tabs.Count-1 downto 0 do
  begin
    if NativeInt(viewBar.Tabs.Objects[i]) = 2 then
    begin
      viewBar.Tabs.Delete(i);
      Break;
    end;
  end;
end;

procedure HookNewEditWindowMenu;
var
  module: TDataModule;
  notify: TNotifyEvent;
begin
  module := TDataModule(Application.FindComponent('EditorActionLists'));
  if module = nil then Exit;

  ecNewEditWindow := TAction(module.FindComponent('ecNewEditWindow'));
  if ecNewEditWindow = nil then Exit;

  TMethod(notify).Code := @AltNewEditWindowExecute;
  TMethod(notify).Data := nil;

  Org_ecNewEditWindowExecute := ecNewEditWindow.OnExecute;
  ecNewEditWindow.OnExecute := notify;
end;

procedure UnhookNewEditWindowMenu;
begin
  if ecNewEditWindow = nil then Exit;

  ecNewEditWindow.OnExecute := Org_ecNewEditWindowExecute;
end;

type
  PFindModuleRec = ^TFindModuleRec;
  TFindModuleRec = record
    Path: string;
    RegEx: TRegEx;
    Result: string;
  end;

function EnumModuleFunc(HInstance: NativeInt; Data: Pointer): Boolean;
var
  name: string;
begin
  name := GetModuleName(HInstance);
  if name.StartsWith(PFindModuleRec(Data)^.Path, True) and PFindModuleRec(Data).RegEx.Match(ExtractFileName(name)).Success then
  begin
    PFindModuleRec(Data)^.Result := ExtractFileName(name);
    Result := False;
  end
  else begin
    Result := True;
  end;
end;

function GetCoreideFileName: string;
var
  rec: TFindModuleRec;
begin
  rec.Path := ExtractFilePath(ParamStr(0));
  rec.RegEx := TRegEx.Create('^coreide[0-9]+\.bpl$', [roIgnoreCase, roCompiled]);
  rec.Result := '';
  EnumModules(EnumModuleFunc, @rec);
  Result := rec.Result;
end;

procedure RegisterEditorViewsChangedEvent;
var
  moduleName: string;
  hModule: THandle;
  notify: TNotifyEvent;
begin
  moduleName := GetCoreideFileName;
  if moduleName = '' then Exit;
  hModule := GetModuleHandle(PChar(moduleName));
  if hModule = 0 then Exit;

  evEditorViewsChangedEvent := GetProcAddress(hModule, sEditorViewsChangedEvent);
  if evEditorViewsChangedEvent = nil then Exit;

  TMethod(notify).Code := @MyEditorViewsChanged;
  TMethod(notify).Data := nil;
  evEditorViewsChangedEvent.Add(notify);
end;

procedure UnregisterEditorViewsChangedEvent;
var
  notify: TNotifyEvent;
begin
  if evEditorViewsChangedEvent = nil then Exit;

  TMethod(notify).Code := @MyEditorViewsChanged;
  TMethod(notify).Data := nil;
  evEditorViewsChangedEvent.Remove(notify);
end;

procedure Register;
begin
  HookNewEditWindowMenu;
  RegisterEditorViewsChangedEvent;
end;

initialization
finalization
  UnhookNewEditWindowMenu;
  UnregisterEditorViewsChangedEvent;
end.
